#include "../SDK/foobar2000.h"

typedef unsigned int    uint32_t;
typedef signed int      int32_t;
typedef unsigned short  uint16_t;
typedef signed short    int16_t;
typedef unsigned char   uint8_t;
typedef signed char     int8_t;

extern "C" {
#include "dts.h"
}

#define BUFFER_SIZE 24576
#define HEADER_SIZE 14
#define FRAME_SAMPLES 256

//extern cfg_int cfg_drc;

class packet_decoder_dts : public packet_decoder {
    bool init()
    {
        cleanup();

        state = dts_init ( 0 );
        if ( !state ) {
            console::error ( "Failed to initialize DTS decoder" );
	        return false;
        }

        return true;
    }

    void cleanup()
    {
        if ( state ) {
            dts_free ( state );
            state = NULL;
        }

        bufptr = buf;
        bufpos = buf + HEADER_SIZE;
        dts_flags = nch = srate = bitrate = 0;
    }

    int get_channel_count ( int flags )
    {
        int channels = 0, tmp = flags & DTS_CHANNEL_MASK;

        if ( tmp > DTS_CHANNEL_MAX ) return 0;

        if ( tmp == DTS_MONO ) channels = 1;
        else if ( tmp < DTS_3F ) channels = 2;
        else if ( tmp < DTS_3F1R ) channels = 3;
        else if ( tmp < DTS_3F2R ) channels = 4;
        else if ( tmp < DTS_4F2R ) channels = 5;
        else if ( tmp == DTS_4F2R ) channels = 6;

        if ( channels && (flags & DTS_LFE) ) channels++;

        return channels;
    }

    unsigned int get_channel_map_mask ( int flags )
    {
        static const unsigned int output_nch[10] = {
            audio_chunk::channel_front_center,                                  // DTS_MONO
            audio_chunk::channel_front_left | audio_chunk::channel_front_right, // DTS_CHANNEL
            audio_chunk::channel_front_left | audio_chunk::channel_front_right, // DTS_STEREO
            audio_chunk::channel_front_left | audio_chunk::channel_front_right, // DTS_STEREO_SUMDIFF
            audio_chunk::channel_front_left | audio_chunk::channel_front_right, // DTS_STEREO_TOTAL
            audio_chunk::channel_front_left | audio_chunk::channel_front_right | audio_chunk::channel_front_center, // DTS_3F
            audio_chunk::channel_front_left | audio_chunk::channel_front_right | audio_chunk::channel_back_center,  // DTS_2F1R
            audio_chunk::channel_front_left | audio_chunk::channel_front_right | audio_chunk::channel_front_center | audio_chunk::channel_back_center, // DTS_3F1R
            audio_chunk::channel_front_left | audio_chunk::channel_front_right | audio_chunk::channel_back_left    | audio_chunk::channel_back_right,  // DTS_2F2R
            audio_chunk::channel_front_left | audio_chunk::channel_front_right | audio_chunk::channel_front_center | audio_chunk::channel_back_left | audio_chunk::channel_back_right, // DTS_3F2R
        };

        unsigned int map = output_nch[flags & DTS_CHANNEL_MASK];
        if (flags & DTS_LFE) map |= audio_chunk::channel_lfe;
        return map;
    }

#if 0
    int get_output_channel_count ( int flags )
    {
        static const int output_nch[10] = {
            1,       /* DTS_MONO */
            2,       /* DTS_CHANNEL */
            2,       /* DTS_STEREO */
            2,       /* DTS_STEREO_SUMDIFF */
            2,       /* DTS_STEREO_TOTAL */
            3,       /* DTS_3F */
            9,       /* DTS_2F1R */
            9,       /* DTS_3F1R */
            6,       /* DTS_2F2R */
            6,       /* DTS_3F2R */
        };
        static const int output_nch_lfe[10] = {
            4,       /* DTS_MONO */
            4,       /* DTS_CHANNEL */
            4,       /* DTS_STEREO */
            4,       /* DTS_STEREO_SUMDIFF */
            4,       /* DTS_STEREO_TOTAL */
            4,       /* DTS_3F */
            9,       /* DTS_2F1R */
            9,       /* DTS_3F1R */
            6,       /* DTS_2F2R */
            6,       /* DTS_3F2R */
        };

        if ( flags & DTS_LFE ) {
            return output_nch_lfe[flags & DTS_CHANNEL_MASK];
        } else {
            return output_nch[flags & DTS_CHANNEL_MASK];
        }
    }
#endif

    dts_state_t *state;

    pfc::array_t<audio_sample> output;
    uint8_t buf[BUFFER_SIZE];
    uint8_t *bufptr, *bufpos;

    int disable_dynrng;
    int dts_flags, nch, srate, bitrate, frame_length;
    unsigned int channel_mask;
    //int hack_nch;

public:
    packet_decoder_dts()
    {
        state = 0;
        disable_dynrng = 1;//!cfg_drc;

        cleanup();
    }

    ~packet_decoder_dts()
    {
        cleanup();
    }

    virtual t_size set_stream_property ( const GUID &p_type, t_size p_param1, const void *p_param2, t_size p_param2size )
    {
        return 0;
    }

    virtual void get_info ( file_info &p_info )
    {
        p_info.info_set_int("samplerate", srate);
        p_info.info_set_int("channels", nch);
        p_info.info_set("codec", "DTS");
        p_info.info_set("encoding", "lossy");
    }

    void open ( const GUID &p_owner, bool p_decode, t_size p_param1, const void *p_param2, t_size p_param2size, abort_callback &p_abort )
    {
        const matroska_setup *setup = (const matroska_setup *)p_param2;
        if ( !init() ) throw exception_io_data();
        nch = setup->channels;
        srate = setup->sample_rate;
    }

    static bool g_is_our_setup ( const GUID &p_owner, t_size p_param1, const void *p_param2, t_size p_param2size )
    {
        if ( p_owner == owner_matroska ) {
            if ( p_param2size == sizeof(matroska_setup) ) {
                const matroska_setup *setup = (const matroska_setup *)p_param2;
                if ( !strncmp(setup->codec_id,"A_DTS", 5) ) return true; else return false;
            } else return false;
        } else return false;
    }

    virtual unsigned get_max_frame_dependency() { return 0; }
    virtual double get_max_frame_dependency_time() { return 0; }

    virtual void reset_after_seek() {}

    virtual bool analyze_first_frame_supported() { return false; }

    virtual void analyze_first_frame ( const void *p_buffer, t_size p_bytes, abort_callback &p_abort ) { return; }

    //virtual bool decode ( const void *data, unsigned bytes, audio_chunk *out )
    virtual void decode ( const void *data, t_size bytes, audio_chunk &p_chunk, abort_callback &p_abort )
    {
        //if ( !state ) return false;

        static const int chan_map[10][6] = {
            { 0, 0, 0, 0, 0, 0 },       // DTS_MONO
            { 0, 1, 0, 0, 0, 0 },       // DTS_CHANNEL
            { 0, 1, 0, 0, 0, 0 },       // DTS_STEREO
            { 0, 1, 0, 0, 0, 0 },       // DTS_STEREO_SUMDIFF
            { 0, 1, 0, 0, 0, 0 },       // DTS_STEREO_TOTAL
            { 2, 0, 1, 0, 0, 0 },       // DTS_3F
            { 0, 1, 2, 0, 0, 0 },       // DTS_2F1R
            { 2, 0, 1, 3, 0, 0 },       // DTS_3F1R
            { 0, 1, 2, 3, 0, 0 },       // DTS_2F2R
            { 2, 0, 1, 3, 4, 0 },       // DTS_3F2R
        };
        static const int chan_map_lfe[10][6] = {
            { 0, 1, 0, 0, 0, 0 },       // DTS_MONO
            { 0, 1, 2, 0, 0, 0 },       // DTS_CHANNEL
            { 0, 1, 2, 0, 0, 0 },       // DTS_STEREO
            { 0, 1, 2, 0, 0, 0 },       // DTS_STEREO_SUMDIFF
            { 0, 1, 2, 0, 0, 0 },       // DTS_STEREO_TOTAL
            { 2, 0, 1, 3, 0, 0 },       // DTS_3F
            { 0, 1, 3, 2, 0, 0 },       // DTS_2F1R
            { 2, 0, 1, 4, 3, 0 },       // DTS_3F1R
            { 0, 1, 3, 4, 2, 0 },       // DTS_2F2R
            { 2, 0, 1, 4, 5, 3 },       // DTS_3F2R
        };
        /*
        static const int chan_map_hack[10][6] = {
            { 2, 3, 0, 0, 0, 0 },       // DTS_MONO
            { 0, 1, 3, 0, 0, 0 },       // DTS_CHANNEL
            { 0, 1, 3, 0, 0, 0 },       // DTS_STEREO
            { 0, 1, 3, 0, 0, 0 },       // DTS_STEREO_SUMDIFF
            { 0, 1, 3, 0, 0, 0 },       // DTS_STEREO_TOTAL
            { 2, 0, 1, 3, 0, 0 },       // DTS_3F
            { 0, 1, 8, 3, 0, 0 },       // DTS_2F1R
            { 2, 0, 1, 8, 3, 0 },       // DTS_3F1R
            { 0, 1, 4, 5, 3, 0 },       // DTS_2F2R
            { 2, 0, 1, 4, 5, 3 },       // DTS_3F2R
        };
        */

        uint8_t *start = (uint8_t *)data;
        uint8_t *end = (uint8_t *)data + bytes;
        unsigned int samples = 0;

        while ( 1 ) {
            int len = end - start;
            if ( !len ) break;
            if ( len > bufpos - bufptr) len = bufpos - bufptr;

            memcpy ( bufptr, start, len );
            bufptr += len;
            start += len;

            if ( bufptr == bufpos ) {
                if ( bufpos == buf + HEADER_SIZE ) {
                    int length = dts_syncinfo ( state, buf, &dts_flags, &srate, &bitrate, &frame_length );
                    if ( !length ) {
                        //console::warning ( "skip" );
                        for ( bufptr = buf; bufptr < buf + HEADER_SIZE-1; bufptr++ ) bufptr[0] = bufptr[1];
                        continue;
                    }

                    nch = get_channel_count ( dts_flags );
                    channel_mask = get_channel_map_mask ( dts_flags );
                    //hack_nch = get_output_channel_count ( dts_flags );
                    //if ( !nch ) return false;

                    bufpos = buf + length;
                } else {
                    int n;
                    audio_sample *out = 0;
                    level_t level = 1.0;
                    sample_t bias = 0;

                    dts_flags &= DTS_CHANNEL_MASK | DTS_LFE;
                    if ( dts_frame(state, buf, &dts_flags, &level, bias) ) goto error;
                    if ( disable_dynrng ) dts_dynrng ( state, NULL, NULL );

                    for ( n = 0; n < dts_blocks_num(state); n++ ) {
                        if ( dts_block(state) ) goto error;
                        output.grow_size ( (samples + FRAME_SAMPLES) * /*hack_*/ nch );
                        out = (audio_sample *)output.get_ptr() + samples * /*hack_*/ nch;
                        float *src = dts_samples ( state );
                        if ( src ) {
                            int tmp = dts_flags & (~DTS_LFE);

                            /*
                            if ( hack_nch != nch ) memset ( out, 0, sizeof(audio_sample) * FRAME_SAMPLES * hack_nch );
                            for ( int j = 0; j < nch; j++ ) {
                                for ( int i = 0; i < FRAME_SAMPLES; i++ ) {
                                    out[i * hack_nch + chan_map_hack[tmp][j]] = src[j * FRAME_SAMPLES + i];
                                }
                            }
                            */

                            if ( dts_flags & DTS_LFE ) {
                                for ( int j = 0; j < nch; j++ ) {
                                    for ( int i = 0; i < FRAME_SAMPLES; i++) {
                                        out[i * nch + chan_map_lfe[tmp][j]] = src[j * FRAME_SAMPLES + i];
                                    }
                                }
                            } else {
                                for ( int j = 0; j < nch; j++ ) {
                                    for ( int i = 0; i < FRAME_SAMPLES; i++ ) {
                                        //out[i * nch + chan_map[tmp][j]] = src[j * nch + i];
                                        out[i * nch + chan_map[tmp][j]] = src[j * FRAME_SAMPLES + i];
                                    }
                                }
                            }

                            samples += FRAME_SAMPLES;
                        }
                    }

                    bufptr = buf;
                    bufpos = buf + HEADER_SIZE;
                    continue;

                    error:
                    console::warning ( "error" );
                    bufptr = buf;
                    bufpos = buf + HEADER_SIZE;
                }
            }
        }

        if ( samples > 0 ) {
            p_chunk.set_data ( output.get_ptr(), samples, /*hack_*/ nch, srate, channel_mask );
        }
    }
};

static packet_decoder_factory_t<packet_decoder_dts> g_packet_decoder_dts_factory;
